import std.string : find;
import std.process : execvp;

char[] glob2regexp(char[] glob) {
  char[] result = "^(?:(?:.|\\n)*/)?+";
  bool need_prot = true;
  for(auto i = 0; i < glob.length; ++i) switch(glob[i]) {
    case '^', '$', '{', '(', ')', '+', '|', '.':
      result ~= '\\';
      result ~= glob[i];
      need_prot = false;
      break;
    case '\\':
      if(i+1 == glob.length) {
	// make a regexp which is as broken as the glob
	result ~= '\\';
	return result;
      }
      switch(glob[i+1]) {
      case '\\', '*', '?', '[':
	result ~= glob[i..i+2];
        ++i;
	need_prot = false;
	break;
      default:
	// do nothing, the next iteration will pick that character up
	// and do the right thing if it is something special
	break;
      }
      break;
    case '?':
      if(need_prot) result ~= "(?:(?<!/|^)|(?!\\.))";
      result ~= "[^/]";
      need_prot = false;
      break;
    case '*':
      if(need_prot) result ~= "(?:(?:(?<!/|^)|(?!\\.))";
      result ~= "[^/]*";
      if(need_prot) result ~= ")?";
      break;
    case '[':
      // we don't handle any of ^ [:hurz:] [.hurz.] or similiar
      { bool neg = false;
	auto beg = i+1;
	if(beg < glob.length && glob[beg] == '!') {
	  neg = true;
	  ++beg;
	}
	auto end = beg+1;
	for(; end < glob.length && glob[end] != ']'; ++end) {}
	if(end > glob.length) end = glob.length;
	if(need_prot) result ~= "(?:(?<!/|^)|(?!\\.))";
	result ~= '[';
	if(neg) result ~= '^';
	result ~= glob[beg..end];
	if(end == glob.length)
	  // make a regexp as broken as the glob
	  return result;
	result ~= ']';
	need_prot = false;
	i = end;
      }
      break;
    default:
      result ~= glob[i];
      need_prot = false;
      break;
  }
  result ~= '$';
  return result;
}

int main(char[][] args) {
  auto orig_args = args.dup;
  enum { unknown, orig, perl }
  auto grep = unknown;
 i_loop:
  for(auto i = 1; i < args.length; ++i) {
    if(args[i].length == 0)
      continue i_loop;
    if(args[i].length == 1)
      continue i_loop;
    if(args[i][0] != '-')
      continue i_loop;
    if(args[i][1] != '-') {
      // single character option, possibly bundle
    c_loop:
      for(auto c = 1; c < args[i].length; ++c) switch(args[i][c]) {
	  // options without arguments that are left alone
	  // the following are identical for grap and pcregrep
	case 'H': // force print filename
	case 'L': // files without match
	case 'V': // version
	case 'c': // count
	case 'h': // supress print filename
	case 'i': // ignore case
	case 'l': // files with match
	case 'n': // print line number
	case 'o': // only matching
	case 'q': // quiet
	case 'r': // recurse
	case 's': // no massages
	case 'v': // invert
	case 'w': // whole words
	case 'x': // whole lines
	  // the following are not defined for pcregrep
	  // leave them for pcregrep to complain
	case 'Z': // zero terminated output
	case 'b': // print byte offset
	case 'z': // zero terminated input
	  // nothing to do
	  continue c_loop;

	  // options with arguments that are left alone
	case 'A': // after context
	case 'B': // before context
	case 'C': // context
	case 'D': // devices
	case 'd': // directories
	case 'e': // regexp given
	case 'f': // regexp from file
	  // the following are not defined for pcregrep
	  // leave them for pcregrep to complain
	case 'm': // match count limit
	  // option with argument -- either ignore the rest of this param
	  // or, if there is no rest, ignore next param
	  if(c+1 == args[i].length)
	    ++i;
	  continue i_loop;

	  // options without arguments that are deleted
	case 'I': // binary file don't match
	case 'T': // initial tab
	case 'U': // file is binary
	case 'a': // text, default
	case 'u': // unix byte offsets
	  goto short_munch;

	  // special options
	case 'F': // fixed strings
	case 'E': // extended
	case 'G': // basic
	  if(grep == unknown) {
	    grep = orig;
	    break i_loop;
	  }
	  goto short_munch;
	case 'P': // perl
	  // grep is either perl or unknown here
	  grep = perl;
	  goto short_munch;
	case 'y': // synonym for i
	  args[i] = args[i].dup;
	  args[i][c] = 'i';
	  continue c_loop;
	case 'R': // synonym for r
	  args[i] = args[i].dup;
	  args[i][c] = 'r';
	  continue c_loop;
	case '0', '1', '2', '3', '4', '5', '6', '7', '8', '9': // context
	  // insert a -C
	  { auto cend = c;
	    for(; cend < args[i].length
		  && args[i][cend] >= '0' && args[i][cend] <= '9';
		++cend) {}
	    if(cend < args[i].length)
	      args = args[0..i+1] ~ ["-" ~ args[i][cend..$]] ~ args[i+1..$];
	    args[i] = args[i][0..c] ~ "C" ~ args[i][c..cend];
	    continue i_loop;
	  }

	default:
	  throw new Error("Unknown option -"~args[i][c]);

	// tools
	short_munch:
	  // munch option
	  args[i] = args[i][0..c] ~ args[i][c+1..$];
	  if(args[i].length == 1) {
	    // ate all options in bundle
	    args = args[0..i] ~ args[i+1..$];
	    --i;
	    continue i_loop;
	  }
	  --c;
	  continue c_loop;
      }
      continue i_loop;
    }
    if(args[i].length == 2) // args[i] == "--"
      break i_loop;
    else { // long options
      auto eq = find(args[i], '=');
      bool has_eq = eq != -1;
      if(!has_eq) eq = args[i].length;
      switch(args[i][2..eq]) {
	//options without arguments which are left alone
      case "color":  //may have an argument, but only with =
      case "colour": // same
      case "count":
      case "files-with-matches":
      case "files-without-match":
      case "help":
      case "ignore-case":
      case "invert-match":
      case "line-number":
      case "line-regexp":
      case "no-filename":
      case "no-messages":
      case "only-matching":
      case "quiet":
      case "version":
      case "with-filename":
      case "word-regexp":
	// the following are not defined for pcregrep
	// leave them for pcregrep to complain
      case "byte-offset":
	continue i_loop;

	//options with arguments which are left alone
      case "after-context":
      case "before-context":
      case "context":
      case "devices":
      case "directories":
      case "exclude-dir":
      case "exclude-from":
      case "file":
      case "label":
      case "regexp":
	// the following are not defined for pcregrep
	// leave them for pcregrep to complain
      case "max-count":
	if(!has_eq) ++i;
	continue i_loop;

	//options without arguments which are deleted
      case "binary":
      case "initial-tab":
      case "line-buffered":
      case "mmap":
      case "null":
      case "null-data":
      case "text":
      case "unix-byte-offsets":
	goto long_munch;

	//options with argument wich are deleted
      case "binary-files":
	if(has_eq || i+1 == args.length)
	  args = args[0..i] ~ args[i+1..$];
	else
	  args = args[0..i] ~ args[i+2..$];
	--i;
	continue i_loop;

	//options with special processing
      case "basic-regexp":
      case "extended-regexp":
      case "fixed-strings":
	if(grep == unknown) {
	  grep = orig;
	  break i_loop;
	}
	goto long_munch;
      case "perl-regexp":
	// grep is either perl or unknown here
	grep = perl;
	goto long_munch;
      case "silent": // same as quiet
	args[i] = "--quiet" ~ args[i][eq..$];
	continue i_loop;
      case "exclude":
      case "include":
	if(has_eq)
	  args[i] = args[i][0..eq+1] ~ glob2regexp(args[i][eq+1..$]);
	else if(i+1 < args.length)
	  args[i+1] = glob2regexp(args[i+1]);
	++i;
	continue i_loop;

      default:
	throw new Error("Unknown option "~args[i]);

	//tools
      long_munch:
	args = args[0..i] ~ args[i+1..$];
	--i;
	continue i_loop;
      }
    }
  }

  if(grep == unknown) grep = perl;

  if(grep == perl) {
    args[0] = "pcregrep";
  } else {
    orig_args[0] = "grep";
    args = orig_args;
  }
  //  execvp("printf", ["printf"] ~ ["<%s>\\n"] ~ args);
  execvp(args[0], args);

  assert(0);
}
